import { __awaiter, __decorate } from "tslib";
import { Auth, Connect, dataServerResultFilter, extractFieldDefinitions, MessageType, normaliseCriteria, toFieldMetadata, } from '@genesislcap/foundation-comms';
import { LifecycleMixin } from '@genesislcap/foundation-utils';
import { attr, customElement, DOM } from '@microsoft/fast-element';
import { take } from 'rxjs/operators';
import { getAvailableIndexes, getAvailableIndexFields, logger } from '../utils';
import { GridProBaseDatasource, StreamBaseDatasource } from './base.datasource';
import { getServerSideFilterParamsByFieldType } from './server-side.grid-definitions';
/**
 * The IServerSideDatasource implementation, used for SSRM (Server-Side Row Model) in the grid.
 * @alpha
 */
export class StreamDatasource extends StreamBaseDatasource {
    constructor(options) {
        super();
        /**
         * The number of rows in the grid. THIS DO NOT CONSIDER CRITERIA_MATCH scenarios.
         * @internal
         */
        this.server_ROWS_COUNT = 0;
        /**
         * The number of rows in the grid. THIS IS A LOCAL COUNT CONSIDERING CONSIDER CRITERIA_MATCH scenarios.
         * @internal
         */
        this.client_ROWS_COUNT = 0;
        this.moreRows = false;
        this.calculatedRowsCount = 0;
        this.lastSequenceId = 0;
        this.dataserverStream = options.stream;
        this.resourceName = options.resourceName;
        this.resourceParams = options.resourceParams;
        this.originalCriteriaMatch = options.resourceParams.CRITERIA_MATCH;
        this.resourceIndexes = options.resourceIndexes;
        this.resourceColDefs = options.resourceColDefs;
        this.maxRows = options.maxRows;
        this.maxView = options.maxView;
        this.rowId = options.rowId;
        this.pagination = options.pagination;
    }
    getRows(params) {
        var _a, _b;
        return __awaiter(this, void 0, void 0, function* () {
            const filterModelBeingApplied = params.request.filterModel;
            const filtersAreBeingApplied = Object.keys(filterModelBeingApplied).length > 0;
            const filtersAreGettingCleaned = Object.keys(filterModelBeingApplied).length === 0;
            const filterModelBeingAppliedDiffersFromCurrent = JSON.stringify(filterModelBeingApplied) !== JSON.stringify(this.currentFilterModel);
            if (filterModelBeingAppliedDiffersFromCurrent || filtersAreBeingApplied) {
                if (filtersAreGettingCleaned && this.currentFilterModel) {
                    this.currentFilterModel = null;
                    this.resourceParams.CRITERIA_MATCH = this.originalCriteriaMatch;
                    yield this.restartSubscription(params);
                    return;
                }
                else if (filterModelBeingAppliedDiffersFromCurrent && filtersAreBeingApplied) {
                    this.currentFilterModel = filterModelBeingApplied;
                    this.resourceParams.CRITERIA_MATCH = this.buildCriteriaMatchFromFilters();
                    yield this.restartSubscription(params);
                    return;
                }
            }
            const toBeAppliedSortModel = params.request.sortModel;
            if (toBeAppliedSortModel.length !== ((_a = this.currentSortModel) === null || _a === void 0 ? void 0 : _a.length) ||
                toBeAppliedSortModel.length > 0) {
                if (toBeAppliedSortModel.length === 0 && this.currentSortModel) {
                    this.calculatedRowsCount = 0;
                    this.currentSortModel = null;
                    this.resourceParams.ORDER_BY = null;
                    this.resourceParams.REVERSE = null;
                    yield this.restartSubscription(params);
                    return;
                }
                else if (toBeAppliedSortModel.length > 0) {
                    const coldIdBeingSorted = toBeAppliedSortModel[0].colId; // Not allowing multiple sorts by user
                    const sortTypeBeingApplied = toBeAppliedSortModel[0].sort;
                    const orderByAndToBeSortedColIds = this.getOrderByAndToBeSortedColIds(this.resourceIndexes, coldIdBeingSorted);
                    if (!orderByAndToBeSortedColIds) {
                        this.calculatedRowsCount = 0;
                        const availableIndexes = getAvailableIndexes(this.resourceIndexes);
                        const availableIndexFields = getAvailableIndexFields(this.resourceIndexes);
                        logger.warn('The FIELD/column (${coldIdBeingSorted}) being sorted is not part of an INDEX, required for the [orderBy] operation. See https://learn.genesis.global/docs/database/data-types/index-entities/');
                        logger.debug('Available indexes:', availableIndexes);
                        logger.debug('Columns that can be sorted with the available indexes:', availableIndexFields);
                        (_b = params.api) === null || _b === void 0 ? void 0 : _b.refreshServerSide({ purge: true });
                        params.fail();
                        return;
                    }
                    else if (JSON.stringify(toBeAppliedSortModel) !== JSON.stringify(this.currentSortModel)) {
                        this.currentSortModel = toBeAppliedSortModel;
                        this.resourceParams.ORDER_BY = orderByAndToBeSortedColIds.orderBy;
                        this.resourceParams.REVERSE = sortTypeBeingApplied === 'desc' ? true : false;
                        yield this.restartSubscription(params);
                        return;
                    }
                }
            }
            if (this.moreRows && params.request.startRow >= Number(this.maxRows)) {
                this.connect.getMoreRows(this.streamSourceRef);
            }
            this.dataserverStream.pipe(take(1)).subscribe((result) => {
                var _a;
                const messageType = result.MESSAGE_TYPE;
                if (messageType &&
                    (messageType === MessageType.LOGOFF_ACK || messageType === MessageType.MSG_NACK)) {
                    params.fail();
                    return;
                }
                this.moreRows = result.MORE_ROWS;
                if (result.ROW) {
                    const nextMessage = dataServerResultFilter(result);
                    this.handleCurrentStreamLoad(nextMessage);
                }
                else {
                    this.rowData = new Map();
                    const successRowData = { rowData: [] };
                    successRowData.rowCount = 0;
                    params.success(successRowData);
                    return;
                }
                this.lastSequenceId = result.SEQUENCE_ID;
                if (this.lastSequenceId === 1) {
                    this.streamSourceRef = result.SOURCE_REF;
                    this.server_ROWS_COUNT = (_a = result.ROWS_COUNT) !== null && _a !== void 0 ? _a : this.rowData.size;
                }
                const successRowData = {
                    rowData: Array.from(this.sortMapByNumericKey(this.rowData).values()),
                    rowCount: this.pagination && !this.resourceParams.CRITERIA_MATCH
                        ? Math.min(this.server_ROWS_COUNT, this.maxView)
                        : undefined,
                };
                this.calculatedRowsCount = this.client_ROWS_COUNT;
                if (this.calculatedRowsCount === 0) {
                    this.calculatedRowsCount = params.request.endRow;
                }
                if (params.request.endRow > this.maxView ||
                    params.request.endRow > this.calculatedRowsCount) {
                    successRowData.rowCount = this.calculatedRowsCount;
                }
                else if (params.request.endRow === this.server_ROWS_COUNT) {
                    successRowData.rowCount = this.server_ROWS_COUNT;
                }
                else if (!this.moreRows && this.server_ROWS_COUNT > this.calculatedRowsCount) {
                    successRowData.rowCount = this.calculatedRowsCount;
                }
                params.success(successRowData);
            });
        });
    }
    sortMapByNumericKey(inputMap) {
        // Convert Map to array of key-value pairs
        const sortedArray = Array.from(inputMap.entries()).sort(([key1], [key2]) => +key2 - +key1);
        // Create a new Map from the sorted array
        const sortedMap = new Map(sortedArray);
        return sortedMap;
    }
    dataLogoff() {
        return __awaiter(this, void 0, void 0, function* () {
            const shouldPerformDataLogoff = this.streamSourceRef;
            if (shouldPerformDataLogoff && this.auth.isLoggedIn) {
                yield this.connect.dataLogoff(this.streamSourceRef);
            }
            this.streamSourceRef = undefined;
        });
    }
    restartSubscription(params) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            yield this.dataLogoff();
            this.rowData = new Map();
            this.client_ROWS_COUNT = 0;
            const streamOnMessage = (message) => message;
            const onError = (error) => logger.error(error);
            this.dataserverStream = undefined;
            this.dataserverStream = this.connect.stream(this.resourceName, streamOnMessage, onError, this.resourceParams);
            (_a = params.api) === null || _a === void 0 ? void 0 : _a.refreshServerSide({ purge: true });
            params.fail();
        });
    }
    getOrderByAndToBeSortedColIds(map, searchValue) {
        for (const [key, value] of map.entries()) {
            const searchIndex = value.findIndex((x) => x === searchValue);
            if (searchIndex !== -1) {
                const copy = [...value];
                const toBeSorted = copy.splice(searchIndex, 1);
                return { orderBy: key, toBeSorted };
            }
        }
    }
    handleCurrentStreamLoad(result) {
        var _a;
        if (!result)
            return;
        const rows = new Map();
        (_a = result.inserts) === null || _a === void 0 ? void 0 : _a.forEach((insertData) => {
            rows.set(insertData[this.rowId], insertData);
        });
        this.rowData = rows;
    }
    buildCriteriaMatchFromFilters() {
        this.resourceParams.CRITERIA_MATCH = this.originalCriteriaMatch;
        if (this.resourceParams.CRITERIA_MATCH &&
            typeof this.resourceParams.CRITERIA_MATCH === 'string') {
            this.resourceParams.CRITERIA_MATCH = this.resourceParams.CRITERIA_MATCH.split(';');
        }
        if (!this.resourceParams.CRITERIA_MATCH) {
            this.resourceParams.CRITERIA_MATCH = [];
        }
        const criteria = this.resourceParams.CRITERIA_MATCH.concat(this.criteriaFromFilters());
        return criteria.join(' && ');
    }
    // TODO: update this following https://learn.genesis.global/docs/server/data-server/advanced/#common-expressions
    criteriaFromFilters() {
        const filters = [];
        this.getFiltersByType('text').forEach((k) => {
            // TODO: no official "boolean" filter type, it's a "text"
            if (!this.currentFilterModel[k].filter &&
                (this.currentFilterModel[k].type === 'false' || this.currentFilterModel[k].type === 'true')) {
                filters.push(`${k} == ${this.currentFilterModel[k].type}`);
            }
            else {
                switch (this.currentFilterModel[k].type) {
                    case 'equals':
                        filters.push(`${k} == '${this.currentFilterModel[k].filter}'`);
                        break;
                    case 'contains':
                        filters.push(`Expr.containsIgnoreCase(${k}, '${this.currentFilterModel[k].filter}')`);
                        break;
                }
            }
        });
        this.getFiltersByType('number').forEach((k) => {
            const value = this.currentFilterModel[k].filter;
            switch (this.currentFilterModel[k].type) {
                case 'equals':
                    !isNaN(value) && filters.push(`${k} == ${value}`);
                    break;
                case 'lessThan':
                    !isNaN(value) && filters.push(`${k} < ${value}`);
                    break;
                case 'greaterThan':
                    !isNaN(value) && filters.push(`${k} > ${value}`);
                    break;
            }
        });
        this.getFiltersByType('date').forEach((k) => {
            const dateFrom = this.currentFilterModel[k].dateFrom
                .replace(/-/g, '')
                .replace('T', '-')
                .split(' ')[0];
            switch (this.currentFilterModel[k].type) {
                case 'equals':
                    filters.push(`Expr.dateIsEqual(${k}, '${dateFrom}')`);
                    break;
                case 'lessThan':
                    filters.push(`Expr.dateIsBefore(${k}, '${dateFrom}')`);
                    break;
                case 'greaterThan':
                    filters.push(`Expr.dateIsAfter(${k}, '${dateFrom}')`);
                    break;
                default:
                    break;
            }
        });
        return filters;
    }
    getFiltersByType(filterType) {
        return Object.keys(this.currentFilterModel)
            .filter((k) => this.resourceColDefs.findIndex((o) => o.NAME === k) !== -1)
            .filter((k) => this.currentFilterModel[k].filterType === filterType);
    }
}
__decorate([
    Auth
], StreamDatasource.prototype, "auth", void 0);
__decorate([
    Connect
], StreamDatasource.prototype, "connect", void 0);
const criteriaDelimiter = ';';
const withoutColumnDefs = null;
/**
 * A Genesis Datasource element, for server-side | SSRM-compatible data fetching and used exclusively by the GridPro element.
 * @remarks Only supports Server-Side Row Model. Requires `@ag-grid-enterprise/server-side-row-model` setup and valid AG Grid Enterprise license.
 * @alpha
 */
let GridProServerSideDatasource = class GridProServerSideDatasource extends LifecycleMixin(GridProBaseDatasource) {
    constructor() {
        super(...arguments);
        this.applyFuncName = 'applyServerSideTransaction';
        this.applyAsyncFuncName = 'applyServerSideTransactionAsync';
        this.pagination = false;
    }
    criteriaChanged(oldCriteria, newCriteria) {
        const criteriaIsNotDuplicate = oldCriteria !== normaliseCriteria(newCriteria, criteriaDelimiter);
        if (this.ssrmDatasource && criteriaIsNotDuplicate) {
            this.reloadResourceData();
        }
    }
    /**
     * Resets the grid data while keeping columnDefs and sends a DATA_LOGOFF message if it's a stream.
     * @remarks This is used when the grid is already initialized and we want to reload the data due to a criteria/filter change.
     * @internal
     */
    reloadResourceData() {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            this.ssrmDatasource.dataLogoff();
            this.clearRowData(withoutColumnDefs);
            this.ssrmDatasource = undefined;
            (_a = this.agGrid.gridApi) === null || _a === void 0 ? void 0 : _a.refreshServerSide({ purge: true });
            (_c = (_b = this.agGrid) === null || _b === void 0 ? void 0 : _b.gridApi) === null || _c === void 0 ? void 0 : _c.showLoadingOverlay();
            this.init();
        });
    }
    resourceNameChanged(oldValue, newValue) {
        if (!oldValue || oldValue === newValue)
            return;
        this.restart();
    }
    connectedCallback() {
        super.connectedCallback();
        const shouldRunConnect = this.shouldRunConnect;
        DOM.queueUpdate(() => {
            if (!shouldRunConnect)
                return;
            this.init();
        });
    }
    disconnectedCallback() {
        super.disconnectedCallback();
        const shouldRunDisconnect = this.shouldRunDisconnect;
        DOM.queueUpdate(() => {
            if (!shouldRunDisconnect)
                return;
            this.deinit();
        });
    }
    deepClone() {
        const copy = super.deepClone();
        copy.deferredGridOptions = structuredClone(this.deferredGridOptions);
        copy.deferredColumnStates = structuredClone(this.deferredColumnStates);
        return copy;
    }
    deinit() {
        var _a;
        this.clearRowData();
        this.ssrmDatasource = undefined;
        this.originalFieldDef = undefined;
        (_a = this.agGrid.gridApi) === null || _a === void 0 ? void 0 : _a.refreshServerSide({ purge: true });
    }
    init() {
        if (this.agGrid && this.resourceName && this.connect.isConnected) {
            this.agGrid.gridOptions = Object.assign({ getRowId: (params) => {
                    if (!params.data) {
                        return null;
                    }
                    return params.data[this.rowId];
                }, getRowStyle(params) {
                    var _a, _b;
                    if ((_b = (_a = params.data) === null || _a === void 0 ? void 0 : _a.ROW_REF) === null || _b === void 0 ? void 0 : _b.includes('_deleted')) {
                        return { color: 'grey' };
                    }
                }, cacheBlockSize: this.maxRows, defaultColDef: {
                    filter: true,
                    resizable: true,
                    sortable: true,
                }, pagination: this.pagination, paginationPageSize: this.maxRows, rowBuffer: 0, rowModelType: 'serverSide', suppressServerSideInfiniteScroll: false, suppressMultiSort: true, onFilterChanged: (params) => { }, onSortChanged: (params) => { } }, this.deferredGridOptions);
            this.agGrid.addEventListener('onGridReady', () => __awaiter(this, void 0, void 0, function* () {
                const meta = yield this.connect
                    .getMetadata(this.resourceName)
                    .catch((e) => logger.error(e));
                if (!meta)
                    return false;
                this.indexes = this.getResourceIndexes(meta.INDEXES);
                this.originalFieldDef = extractFieldDefinitions(meta);
                const fieldMetadata = toFieldMetadata(this.originalFieldDef);
                const agColumnDefs = yield this.getAgColumnDefs(fieldMetadata);
                this.agGrid.gridApi.setColumnDefs(agColumnDefs);
                const dataserverStream = yield this.createDataserverStream();
                this.ssrmDatasource = new StreamDatasource({
                    stream: dataserverStream,
                    resourceName: this.resourceName,
                    resourceParams: this.params,
                    resourceIndexes: this.indexes,
                    resourceColDefs: this.originalFieldDef,
                    maxRows: +this.maxRows,
                    maxView: +this.maxView,
                    rowId: this.rowId,
                    pagination: this.pagination,
                });
                this.agGrid.gridApi.setServerSideDatasource(this.ssrmDatasource);
            }), { once: true });
            return;
        }
        logger.warn(`Application not connected, falling back to local columnDefs/rowData`);
    }
    restart() {
        this.deinit();
        this.init();
    }
    clearRowData(withColumnDefs = true) {
        var _a, _b, _c, _d;
        this.rowData = new Map();
        if (withColumnDefs) {
            (_b = (_a = this.agGrid) === null || _a === void 0 ? void 0 : _a.gridApi) === null || _b === void 0 ? void 0 : _b.setColumnDefs([]);
        }
        this.agTransaction = { remove: [] };
        (_c = this.agGrid.gridApi) === null || _c === void 0 ? void 0 : _c.forEachNode((node) => {
            this.agTransaction.remove.push(node.data);
        });
        (_d = this.agGrid.gridApi) === null || _d === void 0 ? void 0 : _d.applyServerSideTransaction(this.agTransaction);
        this.agTransaction = undefined;
    }
    getResourceIndexes(avaialbleIndexes) {
        const resourceIndexesMap = new Map();
        avaialbleIndexes === null || avaialbleIndexes === void 0 ? void 0 : avaialbleIndexes.forEach((index) => {
            resourceIndexesMap.set(index.NAME, index.FIELDS.split(' '));
        });
        return resourceIndexesMap;
    }
    get params() {
        const params = {
            CRITERIA_MATCH: this.criteria,
            FIELDS: this.fields,
            MAX_ROWS: this.maxRows,
            MAX_VIEW: this.maxView,
            MOVING_VIEW: this.movingView,
            REVERSE: this.reverse,
        };
        const keys = Array.from(this.indexes.keys());
        if (this.orderBy) {
            if (keys.findIndex((key) => key === this.orderBy) > -1) {
                params.ORDER_BY = this.orderBy;
            }
            else {
                const validIndexes = keys.map((key) => key);
                logger.warn('Specified [orderBy] index is not valid and will not be used. See https://learn.genesis.global/docs/database/data-types/index-entities/ - Available indexes:', validIndexes);
            }
        }
        logger.debug('Getting configured params:', params);
        return params;
    }
    get rowModel() {
        return this.agGrid.gridApi.getModel();
    }
    createDataserverStream(existingParams = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const streamOnMessage = (message) => __awaiter(this, void 0, void 0, function* () {
                var _a;
                const messageType = message.MESSAGE_TYPE;
                if (messageType === MessageType.QUERY_UPDATE && message.ROW) {
                    const nextMessage = dataServerResultFilter(message);
                    this.agTransaction = { add: [], remove: [], update: [] };
                    this.handleStreamInserts(nextMessage.inserts);
                    // TODO: deletion flow not being handled yet.
                    // this.handleStreamDeletes(nextMessage.deletes);
                    this.handleStreamDeletesTemp(nextMessage.deletes);
                    this.handleStreamUpdates(nextMessage.updates);
                    this.ssrmDatasource.client_ROWS_COUNT += (_a = nextMessage.inserts.length) !== null && _a !== void 0 ? _a : 0;
                    if (message.SEQUENCE_ID > 1) {
                        const updatedInfo = yield this.connect.snapshot(this.resourceName, {
                            MAX_ROWS: 1,
                            MAX_VIEW: 1,
                            CRITERIA_MATCH: this.criteria,
                        });
                        this.ssrmDatasource.server_ROWS_COUNT = updatedInfo.ROWS_COUNT;
                    }
                    this.applyAllAgTransactions();
                }
            });
            const onError = (error) => logger.error(error);
            return this.connect.stream(this.resourceName, streamOnMessage, onError, existingParams !== null && existingParams !== void 0 ? existingParams : this.params);
        });
    }
    handleStreamDeletesTemp(deletes) {
        if (!deletes)
            return;
        deletes.forEach((deleteData) => {
            const rowNode = this.agGrid.gridApi.getRowNode(deleteData.ROW_REF);
            if (rowNode) {
                rowNode.setData(Object.assign(Object.assign({}, rowNode.data), { ROW_REF: deleteData.ROW_REF + '_deleted' }));
            }
        });
    }
    getAgColumnDefs(fieldsMetadata) {
        return __awaiter(this, void 0, void 0, function* () {
            const colDefsFromGenesisData = this.generateColumnDefsFromMetadata(fieldsMetadata, getServerSideFilterParamsByFieldType);
            colDefsFromGenesisData.map((colDef) => {
                return {
                    sortable: [].concat(...this.indexes.values()).includes(colDef.field),
                };
            });
            const colDefsMergedWithTemplateDefs = yield this.agGrid.mergeAllColumnDefsAndStates(colDefsFromGenesisData, true);
            return colDefsMergedWithTemplateDefs;
        });
    }
    applyTransaction(agTransaction) {
        return this.agGrid.gridApi.applyServerSideTransaction(agTransaction);
    }
    applyTransactionAsync(agTransaction, callback) {
        this.agGrid.gridApi.applyServerSideTransactionAsync(agTransaction, callback);
    }
};
__decorate([
    attr({ mode: 'boolean' })
], GridProServerSideDatasource.prototype, "pagination", void 0);
GridProServerSideDatasource = __decorate([
    customElement({
        name: 'grid-pro-server-side-datasource',
    })
], GridProServerSideDatasource);
export { GridProServerSideDatasource };
